# from .core import *
# from .utils import *

import os

from . import tl
from . import pl
from . import io

from .core import lineage, pair, sequence
from .core.sequence import Sequence
from .core.pair import Pair
from .core.lineage import Lineage

from .utils import alignment, cluster, decorators, jobs, log, phylogeny, pipeline, progbar, s3, seqio, utilities
from .utils import alignment as aln
from .utils import pipeline as path

from .version import __version__

BINARY_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), 'bin'))



#!/usr/bin/env python
# filename: batch_cellranger.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from argparse import ArgumentParser
import csv
import os
import shutil
import subprocess as sp
import sys
import time
import urllib

import yaml

from natsort import natsorted

from sample_sheet import SampleSheet

from ..utils import log
from ..utils.pipeline import list_files, make_dir


from ..version import __version__


def parse_arguments(print_help=False):
    parser = ArgumentParser(prog='batch_cellranger', description="Batch CellRanger processing of one or more 10x Genomics samples.")
    parser.add_argument('-p', '--project-directory', dest='project_dir',  required=True,
                        help="The project directory, where run data will be downloaded \
                        and output files will be written. Required.")
    parser.add_argument('-c', '--config-file', dest='config_file', required=True,
                        help="The config file, in YML format. Required.")
    parser.add_argument('-d', '--debug', dest="debug", action='store_true', default=False,
                        help="If set, logs information about successfully assigned sequences as well \
                        as unsuccessful sequences. Useful for debugging small test datasets, \
                        as the logging is fairly detailed. \
                        Default is to only log unsuccessful sequences.")
    parser.add_argument('-v', '--version', action='version', \
                        version='%(prog)s {version}'.format(version=__version__))
    if print_help:
        parser.print_help()
    else:
        args = parser.parse_args()
        args.project_dir = os.path.abspath(args.project_dir)
        return args


class Args():
    def __init__(self, project_dir=None, config_file=None, debug=False):
        super(Args, self).__init__()
        self.project_dir = project_dir
        self.config_file = config_file
        self.debug = debug


class Config():
    '''
    ``Config`` provides the following attributes:

      - ``config_file``: path to the configuration file, in YAML format.
      - ``runs``: a list of ``Run`` objects
      - ``samples``: a list of `Sample` objects. Samples are parsed from `Run`s, so
                   samples may exist in this list that will not be processed by any
                   of the cellranger operations.
      - ``ops``: a dictionary with cellranger operations (count, vdj, aggr or features)
               as keys and a list of subjects as values. Maps the operation with
               the samples on which the operation will be performed.
      - ``reference``: dictionary mapping sample names to the VDJ reference. Must include
                     a ``default`` reference, which will be used for all subjects not
                     specifically named in the dictionary.
      - ``transcriptome``: same as ``reference``, but mapping samples to a reference
                         transcriptome for ``cellranger count`` operations.
      - ``feature_reference``: same as ``reference``, but mapping samples to a 
                             feature reference.
      - ``uiport``: port for the cellranger UI. Default is 72647.
      - ``cellranger``: path to the cellranger binary. Default is "cellranger", which
                      assumes that the cellranger binary is on your PATH.

    '''
    def __init__(self, config_file):
        self.config_file = os.path.abspath(config_file)
        self.reference = None
        self.transcriptome = None
        self.feature_reference = None
        self.uiport = None
        self.cellranger = None
        self._runs = None
        self._samples = None
        self._ops = None
        self._parse_config_file()
        
    def __repr__(self):
        rlist = ['BATCH CELLRANGER CONFIGURATION']
        rlist.append('------------------------------')
        rlist.append('config file: {}'.format(self.config_file))
        rlist.append('VDJ reference:')
        rlist.append('  - default: {}'.format(self.reference['default']))
        for k, v in self.reference.items():
            if k == 'default':
                continue
            rlist.append('  - {}: {}'.format(k, v))
        rlist.append('transcriptome:')
        rlist.append('  - default: {}'.format(self.transcriptome['default']))
        for k, v in self.transcriptome.items():
            if k == 'default':
                continue
            rlist.append('  - {}: {}'.format(k, v))
        rlist.append('feature reference:')
        rlist.append('  - default: {}'.format(self.feature_reference['default']))
        for k, v in self.feature_reference.items():
            if k == 'default':
                continue
            rlist.append('  - {}: {}'.format(k, v))
        rlist.append('UI port: {}'.format(self.uiport))
        rlist.append('cellranger binary: {}'.format(self.cellranger))
        rlist.append('runs: {}'.format([r.name for r in self.runs]))
        rlist.append('samples: {}'.format([s.name for s in self.samples]))
        rlist.append('operations:')
        rlist.append('  - vdj: {}'.format(self.ops.get('vdj', [])))
        rlist.append('  - count: {}'.format(self.ops.get('count', [])))
        # rlist.append('  - features: {}'.format(self.ops.get('features', [])))
        rlist.append('  - aggr:')
        for k, v in self.ops.get('aggr', {}).items():
            rlist.append('    - {}: {}'.format(k, v))
        return '\n'.join(rlist)

    
    @property
    def runs(self):
        if self._runs is None:
            return []
        return self._runs

    @runs.setter
    def runs(self, runs):
        self._runs = runs

    
    @property
    def samples(self):
        if self._samples is None:
            return []
        return self._samples

    @samples.setter
    def samples(self, samples):
        self._samples = samples

    
    @property
    def ops(self):
        if self._ops is None:
            return []
        return self._ops

    @ops.setter
    def ops(self, ops):
        self._ops = ops


    def _parse_config_file(self):
        with open(self.config_file) as f:
            config = yaml.safe_load(f)
        # parse runs
        self.runs = [Run(name, cfg) for name, cfg in config['runs'].items()]
        # collect samples from runs
        samples = []
        for run in self.runs:
            if run.samples is not None:
                samples += run.samples
        self.samples = list(set(samples))
        # # assign runs to each sample:
        # for run in self.runs:
        #     for s in samples:
        #         if s.name in [s.name for s in run.samples]:
        #             s.add_run(run.name)
        # parse ops
        self.ops = {}
        self.ops['vdj'] = config.get('vdj', [])
        self.ops['count'] = config.get('count', {})
        self.ops['aggr'] = config.get('aggr', {})
        # assign ops to each sample
        for op, samples in self.ops.items():
            if op in ['count']:
                samples = [k for subject_dict in samples for k in subject_dict.keys()]
            for s in self.samples:
                if s.name in samples:
                    s.add_op(op)
        # references
        self.reference = config.get('vdj_reference', {})
        self.transcriptome = config.get('transcriptome', {})
        self.feature_reference = config.get('feature_reference', {})
        # assign references/transcriptomes to each sample:
        for s in self.samples:
            s.reference = config['vdj_reference'].get(s.name, config['vdj_reference']['default'])
            s.transcriptome = config['transcriptome'].get(s.name, config['transcriptome']['default'])
            s.feature_reference = config['feature_reference'].get(s.name, config['feature_reference']['default'])
        # general config options
        self.uiport = config.get('uiport', 72647)
        self.cellranger = config.get('cellranger', 'cellranger')


class Run():
    '''
    Object for aggregation of sequencing run information throughput the 10x processing
    '''
    def __init__(self, name, config):
        self.name = name
        self.config = config
        self.url = config.get('url', None)
        self.path = os.path.abspath(config['path']) if 'path' in config else None
        self.is_compressed = config.get('is_compressed', True)
        self.samplesheet = os.path.abspath(config['samplesheet']) if 'samplesheet' in config else None
        self.simple_csv = os.path.abspath(config['simple_csv']) if 'simple_csv' in config else None
        self.copy_to_project = config.get('copy_to_project', False)
        self._fastq_path = None
        self._samples = None

    def __repr__(self):
        rstring = 'RUN: {}'.format(self.name)
        rlist = [rstring]
        rlist.append('-' * len(rstring))
        if self.url is not None:
            rlist.append('url: {}'.format(self.url))
        if self.path is not None:
            rlist.append('path: {}'.format(self.path))
        rlist.append('compressed: {}'.format(self.is_compressed))
        if self.samplesheet is not None:
            rlist.append('samplesheet: {}'.format(self.samplesheet))
        if self.simple_csv is not None:
            rlist.append('simple csv: {}'.format(self.simple_csv))
        rlist.append('fastq path: {}'.format(self.fastq_path))
        rlist.append('samples: {}'.format(self.samples))
        return '\n'.join(rlist)

    
    @property
    def sample_names(self):
        if self.samples is not None:
            return [s.name for s in self.samples]
        return []


    @property
    def fastq_path(self):
        return self._fastq_path

    @fastq_path.setter
    def fastq_path(self, path):
        self._fastq_path = path


    @property
    def samples(self):
        if self._samples is None:
            self._samples = self._parse_samples()
        return self._samples

    @samples.setter
    def samples(self, samples):
        self._samples = samples


    def print_splash(self):
        l = len(self.name)
        logger.info('')
        logger.info('-' * (l + 4))
        logger.info('  ' + self.name)
        logger.info('-' * (l + 4))


    def get(self, raw_dir, log_dir=None, debug=None):
        destination = os.path.join(os.path.abspath(raw_dir), self.name)
        if all([self.path is not None, self.copy_to_project, not self.is_compressed]):
            self.path = self._copy(destination, log_dir=log_dir, debug=debug)
        if self.url is not None:
            self.path = self._download(self.url, destination, log_dir=log_dir, debug=debug)
        if self.is_compressed:
            self.path = self._decompress(self.path, destination, log_dir=log_dir, debug=debug)


    def mkfastq(self, fastq_dir, cellranger='cellranger', uiport=None, log_dir=None, debug=None):
        logger.info('Running mkfastq....')
        fastq_dir = os.path.abspath(fastq_dir)
        make_dir(fastq_dir)
        mkfastq_cmd = "cd '{}' && {} mkfastq".format(fastq_dir, cellranger)
        mkfastq_cmd += ' --id={}'.format(self.name)
        mkfastq_cmd += " --run='{}'".format(self.path)
        if self.samplesheet is not None:
            mkfastq_cmd += " --samplesheet='{}'".format(self.samplesheet)
        else:
            mkfastq_cmd += " --csv='{}'".format(self.simple_csv)
        p = sp.Popen(mkfastq_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
        if uiport is not None:
            mkfastq_cmd += " --uiport={}".format(uiport)
        p = sp.Popen(mkfastq_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
        time.sleep(5)
        uifile = os.path.join(fastq_dir, '{}/_uiport'.format(self.name))
        with open(uifile) as f:
            uistring = f.read().strip()
        external_ip = urllib.request.urlopen('https://api.ipify.org').read().decode('utf8')
        uistring = 'http://' + external_ip + ':' + uistring.split(':')[-1]
        logger.info('  - UI is at {}'.format(uistring))
        o, e = p.communicate()
        if debug:
            logger.info('\nMKFASTQ')
            logger.info(mkfastq_cmd)
            logger.info(o)
            logger.info(e)
        if log_dir is not None:
            log_subdir = os.path.join(log_dir, 'mkfastq')
            make_dir(log_subdir)
            write_log(self.name, log_subdir, stdout=o, stderr=e)
        # logger.info('done')
        ## NEED TO DOUBLE-CHECK WHAT THE FASTQ PATH ACTUALLY IS
        ## is it just --output-dir? or do they go into an --id subfolder?
        return os.path.join(fastq_dir, '{}/outs/fastq_path'.format(self.name))

    
    def _copy(self, destination, log_dir=None, debug=False):
        shutil.copytree(self.path, destination)
        return destination
    
    
    def _download(self, url, destination, log_dir=None, debug=False):
        logger.info('Downloading run data....')
        destination = os.path.abspath(destination)
        make_dir(destination)
        wget_cmd = "wget {} '{}'".format(url, destination)
        p = sp.Popen(wget_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
        o, e = p.communicate()
        if debug:
            logger.info('\nDOWNLOAD')
            logger.info(wget_cmd)
            logger.info(o)
            logger.info(e)
        if log_dir is not None:
            log_subdir = os.path.join(log_dir, 'download')
            make_dir(log_subdir)
            write_log(self.name, log_subdir, stdout=o, stderr=e)
        fname = os.path.basename(url)
        # logger.info('done')
        return os.path.join(destination, fname)


    def _decompress(self, source, destination, log_dir=None, debug=False):
        logger.info('Decompressing run data....')
        source = os.path.abspath(source)
        destination = os.path.abspath(destination)
        make_dir(destination)
        tar_cmd = "tar xzvf '{}' -C '{}'".format(source, destination)
        p = sp.Popen(tar_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
        o, e = p.communicate()
        if debug:
            logger.info('\nDECOMPRESS')
            logger.info(tar_cmd)
            logger.info(o)
            logger.info(e)
        if log_dir is not None:
            log_subdir = os.path.join(log_dir, 'decompress')
            make_dir(log_subdir)
            write_log(self.name, log_subdir, stdout=o, stderr=e)
        run_dir = destination
        for (root, subdirs, files) in os.walk(destination):
            if 'RTAComplete.txt' in files:
                run_dir = os.path.join(destination, root)
                break
        # logger.info('done')
        return run_dir

    
    def _parse_samples(self):
        if self.samplesheet is not None:
            return self._parse_samplesheet()
        if self.simple_csv is not None:
            return self._parse_simple_csv()

    
    def _parse_samplesheet(self):
        ss = SampleSheet(self.samplesheet)
        samples = []
        for s in ss.samples:
            samples.append(Sample(s.Sample_ID, name=s.Sample_Name, index=s.index))
        return samples

    
    def _parse_simple_csv(self):
        samples = []
        with open(self.simple_csv) as csvfile:
            reader = csv.DictReader(csvfile)
            for r in reader:
                samples.append(Sample(r['Sample'], index=r['Index']))
        return samples



class Sample():
    '''
    Object for aggregating information about a single sample
    '''
    def __init__(self, id, ops=None, name=None, reference=None, transcriptome=None, op_type=None, feature_reference=None, index=None):
        self.id = id
        self.name = name if name is not None else id
        self.index = index
        self.reference = reference
        self.transcriptome = transcriptome
        self.op_type = op_type
        self.feature_reference = feature_reference
        self.vdj_path = None
        self.count_path = None
        self.feature_path = None
        self.aggr_path = None
        self._ops = ops
        self._fastqs = None
        self._runs = None

    def __lt__(self, other):
        return all([self.name < other.name])

    def __hash__(self):
        return hash(self.id)


    @property
    def runs(self):
        if self._runs is None:
            return []
        return self._runs

    @property
    def fastqs(self):
        if self._fastqs is None:
            return []
        return self._fastqs

    @property
    def ops(self):
        if self._ops is None:
            return []
        return self._ops


    def add_run(self, run):
        if self._runs is None:
            self._runs = [run, ]
        else:
            self._runs.append(run)

    def add_fastq_path(self, fastq):
        if self._fastqs is None:
            self._fastqs = [fastq, ]
        else:
            self._fastqs.append(fastq)

    def add_op(self, op):
        if self._ops is None:
            self._ops = [op, ]
        else:
            self._ops.append(op)



#==================
#    OPERATIONS
#==================

def cellranger_vdj(sample, vdj_dir, cellranger='cellranger', uiport=None, log_dir=None, debug=False):
    '''
    docstring
    '''
    vdj_dir = os.path.abspath(vdj_dir)
    vdj_cmd = "cd '{}'".format(vdj_dir)
    vdj_cmd += " && {} vdj --id {} --sample {} --reference '{}'".format(cellranger,
                                                                        sample.name,
                                                                        sample.id,
                                                                        sample.reference)
    for fastq in sample.fastqs:
        vdj_cmd += " --fastq '{}'".format(fastq)
    if uiport is not None:
        vdj_cmd += ' --uiport {}'.format(uiport)
    p = sp.Popen(vdj_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    time.sleep(3)
    uifile = os.path.join(vdj_dir, '{}/_uiport'.format(self.name))
    with open(uifile) as f:
        uistring = f.read().strip()
    external_ip = urllib.request.urlopen('https://api.ipify.org').read().decode('utf8')
    uistring = 'http://' + external_ip + ':' + uistring.split(':')[-1]
    logger.info('CellRanger UI is at {}'.format(uistring))
    o, e = p.communicate()
    if debug:
        logger.info('\nCELLRANGER VDJ')
        logger.info(o)
        logger.info(e)
    if log_dir is not None:
        log_subdir = os.path.join(log_dir, 'vdj')
        make_dir(log_subdir)
        write_log(sample.name, log_subdir, stdout=o, stderr=e)
    return os.path.join(vdj_dir, sample.name)



def cellranger_count(group, samples, feature_ref, count_dir,
                     cellranger='cellranger', uiport=None, log_dir=None, debug=False):
    '''
    docstring
    '''
    count_dir = os.path.abspath(count_dir)
    lib_csv = _make_feature_library_csv(samples, group, count_dir)
    count_cmd = "cd '{}'".format(count_dir)
    count_cmd += " && {} count --id {} --libraries {} --feature_ref {} --transcriptome '{}'".format(cellranger,
                                                                                                    lib_csv,
                                                                                                    feature_ref,
                                                                                                    sample.id,
                                                                                                    sample.transcriptome)
    for fastq in sample.fastqs:
        count_cmd += " --fastqs '{}'".format(fastq)
    if uiport is not None:
        count_cmd += " --uiport '{}'".format(uiport)
    p = sp.Popen(count_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    time.sleep(3)
    uifile = os.path.join(count_dir, '{}/_uiport'.format(self.name))
    with open(uifile) as f:
        uistring = f.read().strip()
    external_ip = urllib.request.urlopen('https://api.ipify.org').read().decode('utf8')
    uistring = 'http://' + external_ip + ':' + uistring.split(':')[-1]
    logger.info('CellRanger UI is at {}'.format(uistring))
    o, e = p.communicate()
    if debug:
        logger.info('\nCELLRANGER COUNT')
        logger.info(o)
        logger.info(e)
    if log_dir is not None:
        log_subdir = os.path.join(log_dir, 'count')
        make_dir(log_subdir)
        write_log(sample.name, log_subdir, stdout=o, stderr=e)
    return os.path.join(count_dir, sample.name)


# def cellranger_feature_barcoding(sample, feature_dir, cellranger='cellranger', uiport=None, log_dir=None, debug=False):
#     feature_dir = os.path.abspath(feature_dir)
#     lib_csv = _make_feature_library_csv(sample, feature_dir)
#     feature_cmd = "cd '{}'".format(feature_dir)
#     feature_cmd += " && {} count --id {} --libraries '{}' --feature-ref '{}' --sample {}'.format(cellranger, 
#                                                                                                  sample.name,
#                                                                                                  lib_csv,
#                                                                                                  sample.feature_reference,
#                                                                                                  sample.name)
#     p = sp.Popen(feature_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     time.sleep(3)
#     uifile = os.path.join(feature_dir, '{}/_uiport'.format(self.name))
#     with open(uifile) as f:
#         uistring = f.read().strip()
#     external_ip = urllib.request.urlopen('https://api.ipify.org').read().decode('utf8')
#     uistring = 'http://' + external_ip + ':' + uistring.split(':')[-1]
#     logger.info('CellRanger UI is at {}'.format(uistring))
#     o, e = p.communicate()
#     if debug:
#         logger.info('\nCELLRANGER FEATURES')
#         logger.info(o)
#         logger.info(e)
#     if log_dir is not None:
#         log_subdir = os.path.join(log_dir, 'features')
#         make_dir(log_subdir)
#         write_log(sample.name, log_subdir, stdout=o, stderr=e)
#     return os.path.join(feature_dir, sample.name)


def _make_feature_library_csv(samples, feature_dir):
    lib_str = 'fastqs,sample,library_type\n'
    for sample in samples:
        for fastq in sample.fastqs:
            lib_str += '{},{},{}'.format(fastq, sample.name, sample.op_type)
    lib_path = os.path.join(feature_dir, '{}_feature-library.csv'.format(sample.name))
    with open(lib_path, 'w') as f:
        f.write(lib_str)
    return lib_path


def cellranger_aggr(samples, group, aggr_dir, normalize='mapped', cellranger='cellranger', uiport=None, log_dir=None, debug=False):
    aggr_dir = os.path.abspath(aggr_dir)
    aggr_csv = _make_aggr_csv(samples, aggr_dir)
    aggr_cmd = "cd '{}'".format(aggr_dir)
    aggr_cmd += " && {} count --id {} --csv '{}' --normalize {}".format(cellranger, 
                                                                        group,
                                                                        aggr_csv,
                                                                        normalize)
    ## Eventually want to replace grabbing stdout/stderr with p.communicate(), so we can grab the standard output
    ## in real time, parse out the url for the UI and print to screen so the user can follow along with the UI
    p = sp.Popen(aggr_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    o, e = p.communicate()
    if debug:
        logger.info('\nCELLRANGER AGGR')
        logger.info(o)
        logger.info(e)
    if log_dir is not None:
        log_subdir = os.path.join(log_dir, 'aggr')
        make_dir(log_subdir)
        write_log(group, log_subdir, stdout=o, stderr=e)
    return os.path.join(aggr_dir, group)


def _make_aggr_csv(samples, aggr_dir):
    aggr_dir = os.path.join(aggr_dir)
    aggr_csv = os.path.join(aggr_dir, 'aggr.csv')
    lines = ['library_id,molecule_h5', ]
    for sample in samples:
        h5_path = os.path.join(sample.count_path, 'outs/molecule_info.h5')
        lines.append('{},{}'.format(sample.id, h5_path))
    with open(aggr_csv, 'w') as f:
        f.write('\n'.join(lines))
    return aggr_csv






def build_directory_structure(cfg):
    dirs = {}
    project_dir = os.path.abspath(args.project_dir)
    make_dir(project_dir)
    shutil.copy(cfg.config_file, os.path.join(project_dir, 'config.yaml'))
    dirs['raw'] = os.path.join(project_dir, 'raw_data')
    dirs['fastq'] = os.path.join(project_dir, 'fastqs')
    dirs['vdj'] = os.path.join(project_dir, 'vdj')
    dirs['count'] = os.path.join(project_dir, 'count')
    # dirs['features'] = os.path.join(project_dir, 'features')
    dirs['aggr'] = os.path.join(project_dir, 'aggr')
    for op in cfg.ops.keys():
        make_dir(dirs[op])
    dirs['log'] = os.path.join(project_dir, 'logs')
    make_dir(dirs['log'])
    return dirs


def write_log(prefix, dir, stdout=None, stderr=None):
    if stdout is not None:
        stdout_file = os.path.join(dir, '{}.stdout'.format(prefix))
        with open(stdout_file, 'w') as f:
            f.write(stdout)
    if stderr is not None:
        stderr_file = os.path.join(dir, '{}.stderr'.format(prefix))
        with open(stderr_file, 'w') as f:
            f.write(stderr)


def print_plan(cfg):
    '''
    prints the plan (runs, samples, ops, references, etc)
    '''
    pass


def print_op_splash(op, samples):
    pass


def print_aggr_splash(aggr):
    pass






def main(args):
    # parse the config file
    cfg = Config(args.config_file)
    print_plan(cfg)

    # build directory structure
    dirs = build_directory_structure(cfg)

    # setup logging
    run_log = os.path.join(dirs['log'], 'run.log')
    log.setup_logging(run_log, print_log_location=False, debug=args.debug)
    global logger
    logger = log.get_logger()

    # mkfastq
    for run in cfg.runs:
        run.print_splash()
        run.get(dirs['raw'], log_dir=dirs['log'], debug=args.debug)
        fastq_path = run.mkfastq(dirs['fastq'],
                                 cellranger=args.cellranger,
                                 log_dir=dirs['log'],
                                 debug=args.debug)
        for sample in cfg.samples:
            if sample.name in run.samples:
                sample.add_fastq_path(fastq_path)

    # # operations (except aggr)
    # opmap = {'vdj': cellranger_vdj,
    #          'count': cellranger_count,
    #          'features': cellranger_feature_barcoding}
    
    # for op in ['vdj', 'count', 'features']:
    #     print_op_splash(op)
    #     opfunction = opmap[op]
    #     for sample in cfg.samples:
    #         if op not in sample.ops:
    #             continue
    #         opfunction(sample,
    #                    dirs[op],
    #                    cellranger=cfg.cellranger,
    #                    uiport=cfg.uiport,
    #                    log_dir=dirs['log'],
    #                    debug=args.debug)
    
    # vdj
    print_op_splash('vdj', cfg.samples)
    for sample in cfg.samples:
        if 'vdj' not in sample.ops:
            continue
        path = cellranger_vdj(sample,
                              dirs['vdj'],
                              cellranger=cfg.cellranger,
                              uiport=cfg.uiport,
                              log_dir=dirs['log'],
                              debug=args.debug)
        sample.vdj_path = path

    # count
    print_op_splash('count', cfg.samples)
    for group, sample_dict in cfg.ops['count']:
        samples = [s for s in cfg.samples if s.name in sample_dict]
        for s in samples:
            s.op_type = sample_dict[s.name]
        path = cellranger_count(samples,
                                dirs['count'],
                                cellranger=cfg.cellranger,
                                uiport=cfg.uiport,
                                log_dir=dirs['log'],
                                debug=args.debug)    
        





    for sample in cfg.samples:
        if 'count' not in sample.ops:
            continue
        path = cellranger_count(sample,
                                dirs['count'],
                                cellranger=cfg.cellranger,
                                uiport=cfg.uiport,
                                log_dir=dirs['log'],
                                debug=args.debug)
        sample.count_path = path

    # # features
    # print_op_splash('features', cfg.samples)
    # for sample in cfg.samples:
    #     if 'features' not in sample.ops:
    #         continue
    #     path = cellranger_feature_barcoding(sample,
    #                                         dirs['features'],
    #                                         cellranger=cfg.cellranger,
    #                                         uiport=cfg.uiport,
    #                                         log_dir=dirs['log'],
    #                                         debug=args.debug)
    #     sample.feature_path = path

    # aggr
    print_aggr_splash(cfg.ops['aggr'])
    for group, sample_names in cfg.ops['aggr'].items():
        samples = [s for s in cfg.samples if s.name in sample_names]
        path = cellranger_aggr(samples,
                               group,
                               dirs['aggr'],
                               normalize='mapped',
                               cellranger=cfg.cellranger,
                               uiport=cfg.uiport,
                               log_dir=dirs['log'],
                               debug=args.debug)
        for s in samples:
            s.aggr_path = path
    
    # compress



if __name__ == "__main__":
    args = parse_arguments()
    main(args)





#!/usr/bin/env python
# filename: features.py


#
# Copyright (c) 2020 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


__all__ = ['assign_cellhash_groups', 'classify_features',
           'positive_feature_cutoff', 'negative_feature_cutoff']


import numpy as np
import pandas as pd

import matplotlib.pyplot as plt

from sklearn.neighbors import KernelDensity

from scipy.signal import argrelextrema
from scipy.stats import scoreatpercentile


def assign_cellhash_groups(df, hashnames=None, rename=None, log_normalize=False,
                           threshold_minimum=4.0, threshold_maximum=10.0, kde_maximum=15.0, 
                           assignments_only=False, update_also=None, debug=False):
    '''
    Assigns cells to hash groups based on cell hashing data.

    Args:
    -----

        df (pd.DataFrame): Squareform input dataframe, containing cellhash UMI counts. Indexes
            should be cells, columns should be cell hashes.
        
        hashnames (iterable): List of hashnames, which correspond to column names in ``df``. 
            If not provided, all columns in ``df`` will be assumed to be  hashnames.
        
        rename (dict): Dictionary relating hasnhames (column names in ``df``) to the preferred
            group name. For example, if the hashname ``'Cellhash1'`` corresponded to the sample 
            ``'Sample1'``, an example ``rename_dict`` argument would be::

                {'Cellhash1': 'Sample1'}

        log_normalize (bool): If ``True``, data will be log2 normalized before hash grouping. Default
            is ``True``.

        threshold_minimum (float): Minimum acceptable kig2-normalized UMI threshold. Potential 
            thresholds below this value will be ignored. Default is ``4.0``.

        threshold_maximum (float): Maximum acceptable kig2-normalized UMI threshold. Potential 
            thresholds above this value will be ignored. Default is ``10.0``.

        kde_maximum (float): Upper limit of the KDE (in log2-normalized UMI counts). This should 
            be below the maximum number of UMI counts, or else strange results may occur. Default 
            is ``15.0``.

        assignments_only (bool): If ``True``, return a pandas ``Series`` object containing only the 
            group assignment. Suitable for appending to an existing dataframe.

        debug (bool): produces plots and prints intermediate information for debugging. Default is 
            ``False``.
    '''
    hash_df = df.copy()
    # make sure the dataframe is counts + 1
    if 0 in hash_df.values:
        hash_df += 1
    if log_normalize:
        hash_df = hash_df.applymap(np.log2)
    if hashnames is None:
        hashnames = hash_df.columns.values
    if rename is None:
        rename = {c: c for c in hash_df.columns.values}

    thresholds = {}
    for hashname in hashnames:
        if debug:
            print(hashname)
        thresholds[hashname] = positive_feature_cutoff(hash_df[hashname],
                                                       threshold_minimum=threshold_minimum,
                                                       threshold_maximum=threshold_maximum,
                                                       kde_maximum=kde_maximum,
                                                       debug=debug)
    
    if debug:
        print('THRESHOLDS')
        print('----------')
        for hashname in hashnames:
            print(f'{hashname}: {thresholds[hashname]}')

    assignments = []
    for _, row in hash_df[hashnames].iterrows():
        a = [h for h in hashnames if row[h] >= thresholds[h]]
        if len(a) == 1:
            assignment = rename.get(a[0], a[0])
        elif len(a) > 1:
            assignment = 'doublet'
        else:
            assignment = 'unassigned'
        assignments.append(assignment)
    hash_df['assignment'] = assignments
    
    if assignments_only:
        return hash_df['assignment']
    elif update_also is not None:
        update_also = update_also.copy()
        update_also['assignment'] = hash_df['assignment']
        return hash_df, update_also
    else:
        return hash_df


def classify_features(df, positive=None, negative=None,
                      positive_threshold_maximum=10.0, positive_threshold_minimum=4.0, positive_kde_maximum=15.0,
                      negative_threshold_maximum=10.0, negative_threshold_minimum=4.0, negative_kde_maximum=15.0,
                      negative_threshold_denominator=2, classifications_only=False, debug=False):
    '''
    Docstring for classify_features.
    '''
    df = df.copy()
    data = {}
    
    if positive is None:
        positive = []
    if negative is None:
        negative = []
    
    for p in positive:
        if debug:
            print(p)
        cutoff = positive_feature_cutoff(df[p],
                                         threshold_maximum=positive_threshold_maximum,
                                         threshold_minimum=positive_threshold_minimum,
                                         kde_maximum = positive_kde_maximum,
                                         debug=debug)
        data[p] = df[p] >= cutoff
    for n in negative:
        if debug:
            print(n)
        cutoff = negative_feature_cutoff(df[n],
                                         threshold_maximum=negative_threshold_maximum,
                                         threshold_minimum=negative_threshold_minimum,
                                         kde_maximum = negative_kde_maximum,
                                         denominator=negative_threshold_denominator,
                                         debug=debug)
        data[n] = df[n] > cutoff
    
    if classifications_only:
        return pd.DataFrame(data)
    else:
        data = {'{}_classified'.format(k): v for k, v in data.items()}
        return df.join(pd.DataFrame(data))
        
    
def positive_feature_cutoff(vals, threshold_maximum=10.0, threshold_minimum=4.0, kde_maximum=15.0,
                            debug=False, show_cutoff_value=False, cutoff_text='cutoff', debug_figfile=None):
    a = np.array(vals)
    k = _bw_silverman(a)
    kde = KernelDensity(kernel='gaussian', bandwidth=k).fit(a.reshape(-1, 1))
    s = np.linspace(0, kde_maximum, num=int(kde_maximum * 100))
    e = kde.score_samples(s.reshape(-1,1))
    
    all_min, all_max = argrelextrema(e, np.less)[0], argrelextrema(e, np.greater)[0]
    if len(all_min) > 1:
        _all_min = np.array([m for m in all_min if s[m] <= threshold_maximum and s[m] >= threshold_minimum])
        min_vals = zip(_all_min, e[_all_min])
        mi = sorted(min_vals, key=lambda x: x[1])[0][0]
        cutoff = s[mi]
    elif len(all_min) == 1:
        mi = all_min[0]
        cutoff = s[mi]
    else:
        cutoff = None
    if debug:
        if cutoff is not None:
            # plot
            plt.plot(s, e)
            plt.fill_between(s, e, y2=[min(e)] * len(s), alpha=0.1)
            plt.vlines(cutoff, min(e), max(e),
                       colors='k', alpha=0.5, linestyles=':', linewidths=2)
            # text
            text_xadj = 0.025 * (max(s) - min(s))
            cutoff_string = f'{cutoff_text}: {round(cutoff, 3)}' if show_cutoff_value else cutoff_text
            plt.text(cutoff - text_xadj, max(e), cutoff_string, ha='right', va='top', fontsize=14)
            # style
            ax = plt.gca()
            for spine in ['right', 'top']:
                ax.spines[spine].set_visible(False)
            ax.tick_params(axis='both', labelsize=12)
            ax.set_xlabel('$\mathregular{log_2}$ UMI counts', fontsize=14)
            ax.set_ylabel('kernel density', fontsize=14)
            # save or show
            if debug_figfile is not None:
                plt.tight_layout()
                plt.savefig(debug_figfile)
            else:
                plt.show()
        print('bandwidth: {}'.format(k))
        print('local minima: {}'.format(s[all_min]))
        print('local maxima: {}'.format(s[all_max]))
        if cutoff is not None:
            print('cutoff: {}'.format(cutoff))
        else:
            print('WARNING: no local minima were found, so the threshold could not be calculated.')
        print('\n\n')
    return cutoff


def negative_feature_cutoff(vals, threshold_maximum=10.0, threshold_minimum=4.0, kde_maximum=15.0, denominator=2.0,
                            debug=False, show_cutoff_value=False, cutoff_text='cutoff', debug_figfile=None):
    a = np.array(vals)
    k = _bw_silverman(a)
    kde = KernelDensity(kernel='gaussian', bandwidth=k).fit(a.reshape(-1, 1))
    s = np.linspace(0, kde_maximum, num=int(kde_maximum * 100))
    e = kde.score_samples(s.reshape(-1,1))
    
    all_min, all_max = argrelextrema(e, np.less)[0], argrelextrema(e, np.greater)[0]
    if len(all_min) > 1:
        _all_min = np.array([m for m in all_min if s[m] <= threshold_maximum and s[m] >= threshold_minimum])
        min_vals = zip(_all_min, e[_all_min])
        mi = sorted(min_vals, key=lambda x: x[1])[0][0]
        ma = [m for m in all_max if s[m] < s[mi]][-1]
        # adj = len(all_max) - len(all_min) - 1
        # ma = all_max[max([0, list(all_min).index(mi) - adj])]
        cutoff = s[int((mi + ma) / denominator)]
    elif len(all_min) == 1:
        mi = all_min[0]
        ma = all_max[0]
        cutoff = s[int((mi + ma) / denominator)]
    else:
        cutoff = None
    if debug:
        if cutoff is not None:
            # plot
            plt.plot(s, e)
            plt.fill_between(s, e, y2=[min(e)] * len(s), alpha=0.1)
            plt.vlines(s[mi], min(e), e[mi], colors='k', alpha=0.5, linestyles=':')
            plt.vlines(s[ma], min(e), e[ma], colors='k', alpha=0.5, linestyles=':')
            plt.vlines(cutoff, min(e), max(e),
                       colors='k', alpha=0.5, linestyles=':', linewidths=2)
            # text
            text_ymin = min(e) + (0.025 * (max(e) - min(e)))
            text_xadj = 0.025 * (max(s) - min(s))
            plt.text(s[mi] + text_xadj, text_ymin, 'local\nmin', ha='left', va='bottom', fontsize=12)
            plt.text(s[ma] - text_xadj, text_ymin, 'local\nmax', ha='right', va='bottom', fontsize=12)
            cutoff_string = f'{cutoff_text}: {round(cutoff, 3)}' if show_cutoff_value else cutoff_text
            plt.text(cutoff + text_xadj, max(e), cutoff_string, ha='left', va='top', fontsize=14)
            # style
            ax = plt.gca()
            for spine in ['right', 'top']:
                ax.spines[spine].set_visible(False)
            ax.tick_params(axis='both', labelsize=12)
            ax.set_xlabel('$\mathregular{log_2}$ UMI counts', fontsize=14)
            ax.set_ylabel('kernel density', fontsize=14)
            # save or show
            if debug_figfile is not None:
                plt.tight_layout()
                plt.savefig(debug_figfile)
            else:
                plt.show()
        print('bandwidth: {}'.format(k))
        print('local minima: {}'.format(s[all_min]))
        print('local maxima: {}'.format(s[all_max]))
        if cutoff is not None:
            print('cutoff: {}'.format(cutoff))
        else:
            print('WARNING: no local minima were found, so the threshold could not be calculated.')
        print('\n\n')
    return cutoff

    
def _bw_silverman(x):
    normalize = 1.349
    IQR = (scoreatpercentile(x, 75) - scoreatpercentile(x, 25)) / normalize
    std_dev = np.std(x, axis=0, ddof=1)
    if IQR > 0:
        A = np.minimum(std_dev, IQR)
    else:
        A = std_dev
    n = len(x)
    return .9 * A * n ** (-0.2)







#!/usr/bin/env python
# filename: sequence.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

from collections import OrderedDict
import csv
import json
import operator
import sys
import uuid

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

from ..utils.utilities import nested_dict_lookup

if sys.version_info[0] > 2:
    STR_TYPES = [str, ]
    from functools import reduce
else:
    STR_TYPES = [str, unicode]


class Sequence(object):
    """
    Container for biological (RNA and DNA) sequences.

    ``seq`` can be one of several things:

        1) a raw sequence, as a string

        2) an iterable, formatted as ``[seq_id, sequence]``

        3) a dict, containing at least the sequence ID and a
           sequence. Alternate ``id_key`` and ``seq_key``
           can be provided at instantiation.

        4) a Biopython ``SeqRecord`` object

        5) an abutils ``Sequence`` object

    If ``seq`` is provided as a string, the sequence ID can optionally be
    provided via the ``id`` keyword argument.  If ``seq`` is a string and ``id`` is not provided,
    a random sequence ID will be generated with ``uuid.uuid4()``.

    Quality scores can be supplied with ``qual`` or as part of a ``SeqRecord`` object.
    If providing both a SeqRecord object with quality scores and quality scores
    via ``qual``, the ``qual`` scores will override the SeqRecord quality scores.

    If ``seq`` is a dictionary, typically the result of a MongoDB query, the dictionary
    can be accessed directly from the ``Sequence`` instance (via the ``annotations`` property). To retrive the value
    for ``'junc_aa'`` in the instantiating dictionary, you would simply::

        s = Sequence(dict)
        junc = s['junc_aa']

    If ``seq`` is a dictionary, an optional ``id_key`` and ``seq_key`` can be provided,
    which tells the ``Sequence`` object which field to use to populate ``Sequence.id`` and
    ``Sequence.sequence``. Defaults for both ``id_key`` and ``seq_key`` are ``None``, which 
    results in abutils trying to determine the appropriate key. For ``id_key``, the following 
    keys are tried: ``['seq_id', 'sequence_id']``. For ``seq_key``, the following keys are tried:
    ``['vdj_nt', 'sequence_nt', 'sequence']``. If none of the attempts are successful, the 
    ``Sequence.id`` or ``Sequence.sequence`` attributes will be ``None``.

    Alternately, the ``__getitem__()`` interface can be used to obtain a slice from the
    ``sequence`` attribute. An example of the distinction::

        d = {'name': 'MySequence', 'sequence': 'ATGC'}
        seq = Sequence(d, id_key='name', seq_key='sequence')

        seq['name']  # 'MySequence'
        seq[:2]  # 'AT'

    If the ``Sequence`` is instantiated with a dictionary, calls to ``__contains__()`` will
    return ``True`` if the supplied item is a key in the dictionary. In non-dict instantiations,
    ``__contains__()`` will look in the ``Sequence.sequence`` field directly (essentially a
    motif search). For example::

        dict_seq = Sequence({'seq_id': 'seq1', 'vdj_nt': 'ACGT'})
        'seq_id' in dict_seq  # TRUE
        'ACG' in dict_seq     # FALSE

        str_seq = Sequence('ACGT', id='seq1')
        'seq_id' in str_seq  # FALSE
        'ACG' in str_seq     # TRUE

    .. note::

        When comparing ``Sequence`` objects, they are comsidered equal only if their
        sequences and IDs are identical. This means that two ``Sequence`` objects
        with identical sequences but without user-supplied IDs won't be equal,
        because their IDs will have been randomly generated.
    """
    def __init__(self, seq, id=None, qual=None, id_key=None, seq_key=None):
        super(Sequence, self).__init__()
        self._input_sequence = None
        self._input_id = id
        self._input_qual = qual
        self.id = None
        self.sequence = None
        self.qual = None
        self._annotations = None
        self._fasta = None
        self._fastq = None
        self._reverse_complement = None
        self._strand = None
        self.id_key = id_key
        self.seq_key = seq_key
        self.description = None

        self._process_input(seq, id, qual)


    def __len__(self):
        '''
        int: Returns the length of the ``sequence`` attribute
        '''
        return len(self.sequence)

    def __iter__(self):
        '''
        iter: Returns an iterator over the ``sequence`` attribute
        '''
        return iter(self.sequence)

    def __reversed__(self):
        '''
        str: Returns the reverse of the ``sequence`` attribute
        '''
        return ''.join(reversed(self.sequence))

    def __contains__(self, item):
        '''
        bool: If the instance was initialzed with a dictonary (which means
            the ``annotations`` attribute is not empty), ``__contains__(key)``
            will return ``True`` if ``key`` is in ``annotations.keys()``. If ``annotations``
            is an empty dict, indicating instantiation without a dictionary,
            ``__contains__(motif)`` will return True if ``motif`` is in the
            ``sequence attribute.

        '''
        if self.annotations:
            return item in self.annotations.keys()
        return item in self.sequence

    def __getitem__(self, key):
        if type(key) == slice:
            return self.sequence[key]
        elif key in self.annotations.keys():
            return self.annotations.get(key, None)
        elif type(key) == int:
            return self.sequence[key]
        return None

    def __setitem__(self, key, val):
        self.annotations[key] = val

    def __eq__(self, other):
        if all([hasattr(other, 'sequence'), hasattr(other, 'id')]):
            return all([self.sequence == other.sequence, self.id == other.id])
        return False


    @property
    def fasta(self):
        '''
        str: Returns the sequence, as a FASTA-formatted string

        Note: The FASTA string is built using ``Sequence.id`` and ``Sequence.sequence``.
        '''
        if not self._fasta:
            self._fasta = '>{}\n{}'.format(self.id, self.sequence)
        return self._fasta

    @property
    def fastq(self):
        '''
        str: Returns the sequence, as a FASTQ-formatted string

        If ``Sequence.qual`` is ``None``, then ``None`` will be returned instead of a
        FASTQ string
        '''
        if self.qual is None:
            self._fastq = None
        else:
            if self._fastq is None:
                self._fastq = '@{}\n{}\n+\n{}'.format(self.id, self.sequence, self.qual)
        return self._fastq

    @property
    def reverse_complement(self):
        '''
        str: Returns the reverse complement of ``Sequence.sequence``.
        '''
        if self._reverse_complement is None:
            self._reverse_complement = self._get_reverse_complement()
        return self._reverse_complement

    @property
    def annotations(self):
        if self._annotations is None:
            self._annotations = {}
        return self._annotations

    @annotations.setter
    def annotations(self, val):
        self._annotations = val

    @property
    def strand(self):
        if self._strand is None:
            self._strand = 'plus'
        return self._strand

    @strand.setter
    def strand(self, strand):
        self._strand = strand


    def as_fasta(self, name_field=None, seq_field=None):
        name = None
        sequence = None
        if name_field is not None:
            name = self.annotations[name_field]
        if seq_field is not None:
            sequence = self.annotations[seq_field]
        if name is None:
            name = self.id
        if sequence is None:
            sequence = self.sequence
        return '>{}\n{}'.format(name, sequence)


    def region(self, start=0, end=None):
        '''
        Returns a region of ``Sequence.sequence``, in FASTA format.

        If called without kwargs, the entire sequence will be returned.

        Args:

            start (int): Start position of the region to be returned. Default
                is 0.

            end (int): End position of the region to be returned. Negative values
                will function as they do when slicing strings.

        Returns:

            str: A region of ``Sequence.sequence``, in FASTA format
        '''
        if end is None:
            end = len(self.sequence)
        return '>{}\n{}'.format(self.id, self.sequence[start:end])


    def keys(self):
        return self.annotations.keys()


    def values(self):
        return self.annotations.values()


    def get(self, key, default=None):
        return self.annotations.get(key, default)


    def _get_reverse_complement(self):
        rc = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A',
              'Y': 'R', 'R': 'Y', 'S': 'S', 'W': 'W',
              'K': 'M', 'M': 'K', 'B': 'V', 'D': 'H',
              'H': 'D', 'V': 'B', 'N': 'N'}
        return ''.join([rc.get(res, res) for res in self.sequence[::-1]])


    def _process_input(self, seq, id, qual):
        if type(seq) in STR_TYPES:
            if id is None:
                id = uuid.uuid4()
            self.sequence = str(seq).upper()
            self.id = id
            self.qual = qual
            self._input_sequence = self.sequence
        elif type(seq) == Sequence:
            self.id = id if id is not None else seq.id
            self.sequence = seq.sequence
            self.description = seq.description
            self.qual = seq.qual
            self._input_sequence = self.sequence
            self._annotations = seq._annotations
        elif type(seq) in [list, tuple]:
            self.id = str(seq[0])
            self.sequence = str(seq[1]).upper()
            self.qual = qual
            self._input_sequence = self.sequence
        elif type(seq) == SeqRecord:
            if qual is None:
                if 'phred_quality' in seq.letter_annotations:
                    qual = seq.letter_annotations['phred_quality']
                elif 'solexa_quality' in seq.letter_annotations:
                    qual = seq.letter_annotations['solexa_quality']
            self.id = id if id is not None else str(seq.id)
            self.description = str(seq.description)
            self.sequence = str(seq.seq).upper()
            self.qual = qual
            self._input_sequence = self.sequence
        elif type(seq) in [dict, OrderedDict]:
            if self.id_key is None:
                id_options = ['seq_id', 'sequence_id']
                if any([k in seq for k in id_options]):
                    self.id_key = [k for k in id_options if k in seq][0]
            if self.seq_key is None:
                seq_options = ['vdj_nt', 'sequence_nt', 'sequence']
                if any([k in seq for k in seq_options]):
                    self.seq_key = [k for k in seq_options if k in seq][0]
            self.id = id if id is not None else seq.get(self.id_key, None)
            self.sequence = seq.get(self.seq_key, None)
            self.qual = qual
            self._input_sequence = self.sequence
            self._annotations = seq



def read_json(json_file, match=None, fields=None, id_key='seq_id', sequence_key='sequence'):
    if match is None:
        match = {}
    if fields is not None:
        if id_key not in fields:
            fields.append(id_key)
        if sequence_key not in fields:
            fields.append(sequence_key)
    sequences = []
    with open(json_file, 'r') as f:
        for line in f:
            if not line.strip():
                continue
            j = json.loads(line.strip())
            try:
                if all([nested_dict_lookup(j, k.split('.')) == v for k, v in match.items()]):
                    if fields is not None:
                        _fields = [f for f in fields if f in j]
                        j = {f: j[f] for f in _fields}
                    sequences.append(Sequence(j, id_key=id_key, seq_key=sequence_key))
            except KeyError:
                continue
    return sequences



def read_csv(csv_file, delimiter=',', match=None, fields=None, id_key='sequence_id', sequence_key='sequence'):
    if match is None:
        match = {}
    if fields is not None:
        if id_key not in fields:
            fields.append(id_key)
        if sequence_key not in fields:
            fields.append(sequence_key)
    sequences = []
    with open(csv_file, 'r') as f:
        reader = csv.DictReader(f, delimiter=delimiter)
        for r in reader:
            try:
                if all([nested_dict_lookup(j, k.split('.')) == v for k, v in match.items()]):
                    if fields is not None:
                        _fields = [f for f in fields if f in r]
                        r = {f: r[f] for f in _fields}
                    sequences.append(Sequence(r, id_key=id_key, seq_key=sequence_key))
            except KeyError:
                continue
    return sequences


def read_fasta(fasta_file):
    with open(fasta_file) as f:
        sequences = [Sequence(s) for s in SeqIO.parse(f, 'fasta')]
    return sequences


def from_mongodb(db, collection, match=None, project=None, limit=None):
    if match is None:
        match = {}
    if project is None:
        project = {}
    if limit is not None:
        results = db[collection].find(match, project, limit=limit)
    else:
        results = db[collection].find(match, project)
    return [Sequence(r) for r in results]

            








#!/usr/bin/env python
# filename: convert.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

from itertools import chain
import os
from zipfile import ZipFile

from Bio import SeqIO

from .pipeline import list_files, make_dir


def abi_to_fasta(input, output):
    '''
    Converts ABI or AB1 files to FASTA format.

    Args:

         input (str): Path to a file or directory containing abi/ab1 files or
            zip archives of abi/ab1 files

        output (str): Path to a directory for the output FASTA files
    '''
    direcs = [input, ]
    # unzip any zip archives
    zip_files = list_files(input, ['zip'])
    if zip_files:
        direcs.extend(_process_zip_files(zip_files))
    # convert files
    for d in direcs:
        files = list_files(d, ['ab1', 'abi'])
        seqs = [SeqIO.read(open(f, 'rb'), 'abi') for f in files]
        # seqs = list(chain.from_iterable(seqs))
        fastas = ['>{}\n{}'.format(s.id, str(s.seq)) for s in seqs]
        ofile = os.path.basename(os.path.normpath(d)) + '.fasta'
        opath = os.path.join(output, ofile)
        open(opath, 'w').write('\n'.join(fastas))


def _process_zip_files(zip_files):
    out_dirs = []
    for z in zip_files:
        out_path = '.'.join(z.split('.')[:-1])
        make_dir(out_path)
        zhandle = ZipFile(z)
        zhandle.extractall(out_path)
        out_dirs.append(out_path)
    return out_dirs



#!/usr/bin/env python
# filename: alignment.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

from copy import copy, deepcopy
import hashlib
from io import StringIO
import json
import logging
import os
import platform
import subprocess as sp
import sys
import tempfile
import traceback

from skbio.alignment import StripedSmithWaterman

from Bio import AlignIO, pairwise2
from Bio.SeqRecord import SeqRecord

from .pipeline import list_files
from ..core.sequence import Sequence

if sys.version_info[0] > 2:
    STR_TYPES = [str, ]
    import nwalign3 as nw
else:
    STR_TYPES = [str, unicode]
    import nwalign as nw

# from .. import BINARY_DIR


__all__ = ['mafft', 'muscle', 'local_alignment', 'global_alignment', 'dot_alignment', 'SSWAlignment', 'NWAlignment']


# -------------------------------------
#
#     MULTIPLE SEQUENCE ALIGNMENT
#
# -------------------------------------



def mafft(sequences=None, alignment_file=None, fasta=None, fmt='fasta', threads=-1, as_file=False,
          reorder=True, print_stdout=False, print_stderr=False, mafft_bin=None):
    '''
    Performs multiple sequence alignment with MAFFT.

    Args:

        sequences (list): Sequences to be aligned. ``sequences`` can be one of four things:

            1. a FASTA-formatted string

            2. a list of BioPython ``SeqRecord`` objects

            3. a list of abutils ``Sequence`` objects

            4. a list of lists/tuples, of the format ``[sequence_id, sequence]``

        alignment_file (str): Path for the output alignment file. If not supplied,
            a name will be generated using ``tempfile.NamedTemporaryFile()``.

        fasta (str): Path to a FASTA-formatted file of sequences. Used as an
            alternative to ``sequences`` when suppling a FASTA file.

        fmt (str): Format of the alignment. Options are 'fasta', 'phylip', and 'clustal'. Default
            is 'fasta'.

        threads (int): Number of threads for MAFFT to use. Default is ``-1``, which
            results in MAFFT using ``multiprocessing.cpu_count()`` threads.

        as_file (bool): If ``True``, returns a path to the alignment file. If ``False``,
            returns a BioPython ``MultipleSeqAlignment`` object (obtained by calling
            ``Bio.AlignIO.read()`` on the alignment file).
        
        print_stdout (bool): If ``True``, prints MAFFT's standard output. Default is ``False``.

        print_stderr (bool): If ``True``, prints MAFFT's standard error. Default is ``False``.

        mafft_bin (str): Path to MAFFT executable. Default is ``None``, which results in 
            using the default system MAFFT binary.

    Returns:

        Returns a BioPython ``MultipleSeqAlignment`` object, unless ``as_file`` is ``True``,
            in which case the path to the alignment file is returned.

    '''
    if sequences:
        fasta_string = _get_fasta_string(sequences)
        fasta_file = tempfile.NamedTemporaryFile(delete=False)
        fasta_file.close()
        ffile = fasta_file.name
        with open(ffile, 'w') as f:
            f.write(fasta_string)
    elif fasta:
        ffile = fasta
    if alignment_file is None:
        alignment_file = tempfile.NamedTemporaryFile(delete=False).name
    aln_format = ''
    if fmt.lower() == 'clustal':
        aln_format = '--clustalout '
    if fmt.lower() == 'phylip':
        aln_format = '--phylipout '
    if reorder:
        aln_format += '--reorder '
    if mafft_bin is None:
        mafft_bin = 'mafft'
    mafft_cline = '{} --thread {} {}{} > {}'.format(mafft_bin, threads, aln_format, ffile, alignment_file)
    mafft = sp.Popen(str(mafft_cline),
                     stdout=sp.PIPE,
                     stderr=sp.PIPE,
                     universal_newlines=True,
                     shell=True)
    stdout, stderr = mafft.communicate()
    if print_stdout:
        print(mafft_cline)
        print(stdout)
    if print_stderr:
        print(stderr)
    os.unlink(ffile)
    if os.stat(alignment_file).st_size == 0:
        return None
    if as_file:
        return alignment_file
    aln = AlignIO.read(open(alignment_file), fmt)
    os.unlink(alignment_file)
    return aln


def muscle(sequences=None, alignment_file=None, fasta=None,
    fmt='fasta', as_file=False, maxiters=None, diags=False,
    gap_open=None, gap_extend=None, muscle_bin=None, debug=False):
    '''
    Performs multiple sequence alignment with MUSCLE.

    Args:

        sequences (list): Sequences to be aligned. ``sequences`` can be one of four things:

            1. a FASTA-formatted string

            2. a list of BioPython ``SeqRecord`` objects

            3. a list of AbTools ``Sequence`` objects

            4. a list of lists/tuples, of the format ``[sequence_id, sequence]``

        alignment_file (str): Path for the output alignment file. If not supplied,
            a name will be generated using ``tempfile.NamedTemporaryFile()``.

        fasta (str): Path to a FASTA-formatted file of sequences. Used as an
            alternative to ``sequences`` when suppling a FASTA file.

        fmt (str): Format of the alignment. Options are 'fasta' and 'clustal'. Default
            is 'fasta'.

        threads (int): Number of threads (CPU cores) for MUSCLE to use. Default is ``-1``, which
            results in MUSCLE using all available cores.

        as_file (bool): If ``True``, returns a path to the alignment file. If ``False``,
            returns a BioPython ``MultipleSeqAlignment`` object (obtained by calling
            ``Bio.AlignIO.read()`` on the alignment file).

        maxiters (int): Passed directly to MUSCLE using the ``-maxiters`` flag.

        diags (int): Passed directly to MUSCLE using the ``-diags`` flag.

        gap_open (float): Passed directly to MUSCLE using the ``-gapopen`` flag. Ignored
            if ``gap_extend`` is not also provided.

        gap_extend (float): Passed directly to MUSCLE using the ``-gapextend`` flag. Ignored
            if ``gap_open`` is not also provided.

        muscle_bin (str): Path to MUSCLE executable. ``abutils`` includes built-in MUSCLE binaries
            for MacOS and Linux, however, a different MUSCLE binary can be provided. Default is
            ``None``, which results in using the appropriate built-in MUSCLE binary.

    Returns:

        Returns a BioPython ``MultipleSeqAlignment`` object, unless ``as_file`` is ``True``,
            in which case the path to the alignment file is returned.
    '''
    if sequences:
        fasta_string = _get_fasta_string(sequences)
    elif fasta:
        fasta_string = open(fasta, 'r').read()
    if muscle_bin is None:
        mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        muscle_bin = os.path.join(mod_dir, 'bin/muscle_{}'.format(platform.system().lower()))
    aln_format = ''
    if fmt == 'clustal':
        aln_format = ' -clwstrict'
    muscle_cline = '{}{} '.format(muscle_bin, aln_format)
    if maxiters is not None:
        muscle_cline += ' -maxiters {}'.format(maxiters)
    if diags:
        muscle_cline += ' -diags'
    if all([gap_open is not None, gap_extend is not None]):
        muscle_cline += ' -gapopen {} -gapextend {}'.format(gap_open, gap_extend)
    if debug:
        print('muscle binary path:', muscle_bin)
        print('muscle command:', muscle_cline)
    muscle = sp.Popen(str(muscle_cline),
                      stdin=sp.PIPE,
                      stdout=sp.PIPE,
                      stderr=sp.PIPE,
                      universal_newlines=True,
                      shell=True)
    if sys.version_info[0] > 2:
        alignment, stderr = muscle.communicate(input=fasta_string)
    else:
        alignment, stderr = muscle.communicate(input=fasta_string)
        alignment = unicode(alignment, 'utf-8')
        stderr = unicode(stderr, 'utf-8')
    if debug:
        print(stderr)
    aln = AlignIO.read(StringIO(alignment), fmt)
    if as_file:
        if not alignment_file:
            alignment_file = tempfile.NamedTemporaryFile().name
        AlignIO.write(aln, alignment_file, fmt)
        return alignment_file
    return aln


def consensus(aln, name=None, threshold=0.51, ambiguous='N'):
    summary_align = AlignInfo.SummaryInfo(aln)
    consensus = summary_align.gap_consensus(threshold=threshold, ambiguous=ambiguous)
    if name is None:
        name = uuid.uuid4()
    consensus_string = str(consensus).replace('-', '')
    return (name, consensus_string.upper())


def _get_fasta_string(sequences):
    if type(sequences) == str:
        return sequences
    elif all([type(s) == Sequence for s in sequences]):
        return '\n'.join([s.fasta for s in sequences])
    else:
        return '\n'.join([Sequence(s).fasta for s in sequences])
    # elif type(sequences[0]) == SeqRecord:
    #     return '\n'.join(['>{}\n{}'.format(seq.id, str(seq.seq).upper()) for seq in sequences])
    # # elif type(sequences[0]) == Sequence:
    # #     return '\n'.join(['>{}\n{}'.format(seq.id, seq.seq) for seq in sequences])
    # elif type(sequences[0]) in [list, tuple]:
    #     return '\n'.join(['>{}\n{}'.format(seq[0], seq[1]) for seq in sequences])



# ----------------------------
#
#     PAIRWISE ALIGNMENT
#
# ----------------------------



def local_alignment(query, target=None, targets=None, match=3, mismatch=-2,
        gap_open=-5, gap_extend=-2, matrix=None, aa=False, gap_open_penalty=None, gap_extend_penalty=None):
    '''
    Striped Smith-Waterman local pairwise alignment.

    Args:

        query: Query sequence. ``query`` can be one of four things:

            1. a nucleotide or amino acid sequence, as a string

            2. a Biopython ``SeqRecord`` object

            3. an AbTools ``Sequence`` object

            4. a list/tuple of the format ``[seq_id, sequence]``

        target: A single target sequence. ``target`` can be anything that
            ``query`` accepts.

        targets (list): A list of target sequences, to be proccssed iteratively.
            Each element in the ``targets`` list can be anything accepted by
            ``query``.

        match (int): Match score. Should be a positive integer. Default is 3.

        mismatch (int): Mismatch score. Should be a negative integer. Default is -2.

        gap_open (int): Penalty for opening gaps. Should be a negative integer.
            Default is -5.

        gap_extend (int): Penalty for extending gaps. Should be a negative
            integer. Default is -2.

        matrix (str, dict): Alignment scoring matrix. Two options for passing the
            alignment matrix:

            - The name of a built-in matrix. Current options are ``blosum62`` and ``pam250``.

            - A nested dictionary, giving an alignment score for each residue pair. Should be formatted
              such that retrieving the alignment score for A and G is accomplished by::

                matrix['A']['G']

        aa (bool): Must be set to ``True`` if aligning amino acid sequences. Default
            is ``False``.

    Returns:

        If a single target sequence is provided (via ``target``), a single ``SSWAlignment``
        object will be returned. If multiple target sequences are supplied (via ``targets``),
        a list of ``SSWAlignment`` objects will be returned.
    '''
    if aa and not matrix:
        err = 'ERROR: You must supply a scoring matrix for amino acid alignments'
        raise RuntimeError(err)
    if not target and not targets:
        err = 'ERROR: You must supply a target sequence (or sequences).'
        raise RuntimeError(err)
    if target:
        targets = [target, ]
    # to maintain backward compatibility with earlier abutils API
    if gap_open_penalty is not None:
        gap_open = -1 * gap_open_penalty
    if gap_extend_penalty is not None:
        gap_extend = -1 * gap_extend_penalty
    alignments = []
    for t in targets:
        try:
            alignment = SSWAlignment(query=query,
                                     target=t,
                                     match=match,
                                     mismatch=mismatch,
                                     matrix=matrix,
                                     gap_open=-1 * gap_open,
                                     gap_extend=-1 * gap_extend,
                                     aa=aa)
            alignments.append(alignment)
        except IndexError:
            continue
    if len(alignments) == 1:
        return alignments[0]
    return alignments


def local_alignment_biopython(query, target=None, targets=None, match=3, mismatch=-2, matrix=None,
        gap_open=-5, gap_extend=-2, aa=False):
    if not target and not targets:
        err = 'ERROR: You must supply a target sequence (or sequences).'
        raise RuntimeError(err)
    if target:
        targets = [target, ]
    alignments = []
    for t in targets:
        try:
            alignment = alignment = BiopythonAlignment(query=query,
                                                       target=t,
                                                       match=match,
                                                       mismatch=mismatch,
                                                       matrix=matrix,
                                                       gap_open=gap_open,
                                                       gap_extend=gap_extend,
                                                       aa=aa)
            alignments.append(alignment)
        except IndexError:
            continue
    if len(alignments) == 1:
        return alignments[0]
    return alignments


def global_alignment(query, target=None, targets=None, match=3, mismatch=-2, gap_open=-5, gap_extend=-2,
        score_match=None, score_mismatch=None, score_gap_open=None,
        score_gap_extend=None, matrix=None, aa=False):
    '''
    Needleman-Wunch global pairwise alignment.

    With ``global_alignment``, you can score an alignment using different
    paramaters than were used to compute the alignment. This allows you to
    compute pure identity scores (match=1, mismatch=0) on pairs of sequences
    for which those alignment parameters would be unsuitable. For example::

        seq1 = 'ATGCAGC'
        seq2 = 'ATCAAGC'

    using identity scoring params (match=1, all penalties are 0) for both alignment
    and scoring produces the following alignment::

        ATGCA-GC
        || || ||
        AT-CAAGC

    with an alignment score of 6 and an alignment length of 8 (identity = 75%). But
    what if we want to calculate the identity of a gapless alignment? Using::

        global_alignment(seq1, seq2,
                         gap_open=20,
                         score_match=1,
                         score_mismatch=0,
                         score_gap_open=10,
                         score_gap_extend=1)

    we get the following alignment::

        ATGCAGC
        ||  |||
        ATCAAGC

    which has an score of 5 and an alignment length of 7 (identity = 71%). Obviously,
    this is an overly simple example (it would be much easier to force gapless alignment
    by just iterating over each sequence and counting the matches), but there are several
    real-life cases in which different alignment and scoring paramaters are desirable.

    Args:

        query: Query sequence. ``query`` can be one of four things:

            1. a nucleotide or amino acid sequence, as a string

            2. a Biopython ``SeqRecord`` object

            3. an AbTools ``Sequence`` object

            4. a list/tuple of the format ``[seq_id, sequence]``

        target: A single target sequence. ``target`` can be anything that
            ``query`` accepts.

        targets (list): A list of target sequences, to be proccssed iteratively.
            Each element in the ``targets`` list can be anything accepted by
            ``query``.

        match (int): Match score for alignment. Should be a positive integer. Default is 3.

        mismatch (int): Mismatch score for alignment. Should be a negative integer. Default is -2.

        gap_open (int): Penalty for opening gaps in alignment. Should be a negative integer.
            Default is -5.

        gap_extend (int): Penalty for extending gaps in alignment. Should be a negative
            integer. Default is -2.

        score_match (int): Match score for scoring the alignment. Should be a positive integer.
            Default is to use the score from ``match`` or ``matrix``, whichever is provided.

        score_mismatch (int): Mismatch score for scoring the alignment. Should be a negative
            integer. Default is to use the score from ``mismatch`` or ``matrix``, whichever
            is provided.

        score_gap_open (int): Gap open penalty for scoring the alignment. Should be a negative
            integer. Default is to use ``gap_open``.

        score_gap_extend (int): Gap extend penalty for scoring the alignment. Should be a negative
            integer. Default is to use ``gap_extend``.

        matrix (str, dict): Alignment scoring matrix. Two options for passing the alignment matrix:

            - The name of a built-in matrix. Current options are ``blosum62`` and ``pam250``.

            - A nested dictionary, giving an alignment score for each residue pair. Should be
              formatted such that retrieving the alignment score for A and G is accomplished by::

                matrix['A']['G']

        aa (bool): Must be set to ``True`` if aligning amino acid sequences. Default
            is ``False``.

    Returns:

        If a single target sequence is provided (via ``target``), a single ``NWAlignment``
        object will be returned. If multiple target sequences are supplied (via ``targets``),
        a list of ``NWAlignment`` objects will be returned.
    '''
    if not target and not targets:
        err = 'ERROR: You must supply a target sequence (or sequences).'
        raise RuntimeError(err)
    if target:
        targets = [target, ]
    if type(targets) not in (list, tuple):
        err = 'ERROR: ::targets:: requires an iterable (list or tuple).'
        err += 'For a single sequence, use ::target::'
        raise RuntimeError(err)
    alignments = []
    for t in targets:
        alignment = NWAlignment(query=query,
                                target=t,
                                match=match,
                                mismatch=mismatch,
                                gap_open=gap_open,
                                gap_extend=gap_extend,
                                score_match=score_match,
                                score_mismatch=score_mismatch,
                                score_gap_open=score_gap_open,
                                score_gap_extend=score_gap_extend,
                                matrix=matrix,
                                aa=aa)
        alignments.append(alignment)
    if target is not None:
        return alignments[0]
    return alignments



# -------------------------------------
#
#           MODELS
#
# -------------------------------------



class BaseAlignment(object):
    """
    Base class for local and global pairwise alignments.

    .. note::

        All comparisons between ``BaseAlignments``
        are done on the ``score`` attribute (which must be implemented
        by any classes that subclass ``BaseAlignment``). This was done
        so that sorting alignments like so::

            alignments = [list of alignments]
            alignments.sort(reverse=True)

        results in a sorted list of alignments from the highest alignment
        score to the lowest.

    Attributes:

        query (Sequence): The input query sequence, as an AbTools
            ``Sequence`` object.

        target (Sequence): The input target sequence, as an AbTools
            ``Sequence`` object.

        target_id (str): ID of the target sequence.

        raw_query: The raw query, before conversion to a ``Sequence``.

        raw_target: The raw target, before conversion to a ``Sequence``.

    """
    def __init__(self, query, target, matrix,
            match, mismatch, gap_open, gap_extend, aa):
        super(BaseAlignment, self).__init__()
        self.query = self._process_sequence(query, aa=aa)
        self.target = self._process_sequence(target, aa=aa)
        self.raw_query = query
        self.raw_target = target
        self._matrix = matrix
        self._match = int(match)
        self._mismatch = int(mismatch)
        self._gap_open = int(gap_open)
        self._gap_extend = int(gap_extend)
        self._aa = bool(aa)

    def __repr__(self):
        if len(self.aligned_query) > 20:
            qstring = '{}...{}'.format(self.aligned_query[:10], self.aligned_query[-10:])
            mstring = '{}...{}'.format(self.alignment_midline[:10], self.alignment_midline[-10:])
            tstring = '{}...{}'.format(self.aligned_target[:10], self.aligned_target[-10:])
        else:
            qstring = self.aligned_query
            mstring = self.alignment_midline
            tstring = self.aligned_target
        return_string = '\n\n'
        return_string += 'Pairwise Alignment\n'
        return_string += '------------------\n\n'
        return_string += 'query:  {}\n'.format(qstring)
        return_string += '        {}\n'.format(mstring)
        return_string += 'target: {}\n\n'.format(tstring)
        return_string += 'score: {}\n'.format(str(self.score))
        return_string += 'type: {}\n'.format(self.alignment_type)
        return_string += 'length: {}'.format(str(len(self.aligned_query)))
        print(return_string)
        return ''

    def __str__(self):
        return_string = ''
        return_string += '{}\n'.format(self.aligned_query)
        return_string += '{}\n'.format(self.alignment_midline)
        return_string += '{}\n'.format(self.aligned_target)
        return return_string

    def __len__(self):
        return len(self.aligned_query)

    def __eq__(self, other):
        if not hasattr(other, 'score'):
            if type(other) in [int, float]:
                return self.score == other
            return False
        return self.score == other.score

    def __lt__(self, other):
        if not hasattr(other, 'score'):
            if type(other) in [int, float]:
                return self.score == other
            return False
        return self.score < other.score

    def __le__(self, other):
        if not hasattr(other, 'score'):
            if type(other) in [int, float]:
                return self.score == other
            return False
        return self.score <= other.score

    def __gt__(self, other):
        if not hasattr(other, 'score'):
            if type(other) in [int, float]:
                return self.score == other
            return False
        return self.score > other.score

    def __ge__(self, other):
        if not hasattr(other, 'score'):
            if type(other) in [int, float]:
                return self.score == other
            return False
        return self.score >= other.score


    @property
    def target_id(self):
        return self._target_id

    @target_id.setter
    def target_id(self, target_id):
        self._target_id = target_id


    @staticmethod
    def _process_sequence(sequence, aa):
        if type(sequence) == Sequence:
            return sequence
        return Sequence(sequence)

    def _alignment_midline(self):
        midline = ''
        for q, t in zip(self.aligned_query, self.aligned_target):
            if q == t:
                midline += '|'
            else:
                midline += ' '
        return midline


class SSWAlignment(BaseAlignment):
    """
    Structure for performing and analyzing a Smith-Waterman local alignment.

    .. note:

        Exposed attributes and methods are the same as ``NWAlignment``, so
        local and global alignmnts can be handled in the same way. In fact,
        since comparisons are made based on score, local and global alignments
        can be directly compared with constructions like::

            local_aln == global_aln
            local_aln > global_aln
            alignments = sorted([global_aln, local_aln])

    Attributes:

        alignment_type (str): Is 'local' for all ``SSWAlignment`` objects.

        aligned_query (str): The aligned query sequence (including gaps).

        aligned_target (str): The aligned target sequence (including gaps).

        alignment_midline (str): Midline for the aligned sequences, with ``|`` indicating
          matches and a gap indicating mismatches::

            print(aln.aligned_query)
            print(aln.alignment_midline)
            print(aln.aligned_target)

            # ATGC
            # || |
            # ATCC

        score (int): Alignment score.

        query_begin (int): Position in the raw query sequence at which
            the optimal alignment begins.

        query_end (int): Position in the raw query sequence at which the
            optimal alignment ends.

        target_begin (int): Position in the raw target sequence at which
            the optimal alignment begins.

        target_end (int): Position in the raw target sequence at which the
            optimal alignment ends.
    """
    def __init__(self, query, target, match=3, mismatch=-2, matrix=None,
            gap_open=5, gap_extend=2, aa=False):
        super(SSWAlignment, self).__init__(query, target, matrix,
            match, mismatch, gap_open, gap_extend, aa)

        self.alignment_type = 'local'
        self._alignment = self._align()
        self.aligned_query = self._alignment.aligned_query_sequence
        self.aligned_target = self._alignment.aligned_target_sequence
        self.alignment_midline = self._alignment_midline()
        self.score = self._alignment.optimal_alignment_score
        self.cigar = self._alignment.cigar
        self.query_begin = self._alignment.query_begin
        self.query_end = self._alignment.query_end
        self.target_begin = self._alignment.target_begin
        self.target_end = self._alignment.target_end_optimal
        self._alignment = None

    def _align(self):
        if sys.version_info[0] == 2:
            query = self.query.sequence.encode('ascii') if isinstance(self.query.sequence, unicode) else self.query.sequence
            target = self.target.sequence.encode('ascii') if isinstance(self.target.sequence, unicode) else self.target.sequence
        else:
            query = self.query.sequence
            target = self.target.sequence
        aligner = StripedSmithWaterman(query,
                                       match_score=self._match,
                                       mismatch_score=self._mismatch,
                                       gap_open_penalty=self._gap_open,
                                       gap_extend_penalty=self._gap_extend,
                                       substitution_matrix=self._matrix,
                                       protein=self._aa)
        return aligner(target)


class BiopythonAlignment(BaseAlignment):
    def __init__(self, query, target, match=3, mismatch=-2, matrix=None,
            gap_open=5, gap_extend=2, aa=False):
        super(BiopythonAlignment, self).__init__(query, target, matrix,
            match, mismatch, gap_open, gap_extend, aa)

        self.alignment_type = 'local'
        self._aln = self._align()
        aln_query, aln_target, score, begin, end = self._aln
        self.aligned_query = aln_query[begin:end]
        self.aligned_target = aln_target[begin:end]
        self.alignment_midline = self._alignment_midline()
        self.score = score
        self.query_begin = self._get_begin_pos(aln_query, begin)
        self.query_end = self._get_end_pos(aln_query, end)
        self.target_begin = self._get_begin_pos(aln_target, begin)
        self.target_end = self._get_end_pos(aln_target, end)

    def _align(self):
        aln = pairwise2.align.localms(self.query.sequence,
                                      self.target.sequence,
                                      self._match,
                                      self._mismatch,
                                      self._gap_open,
                                      self._gap_extend)
        return aln[0]


    def _get_begin_pos(self, seq, begin):
        dashes = seq.count('-', 0, begin)
        return begin - dashes


    def _get_end_pos(self, seq, end):
        return len(seq[:end].replace('-', ''))


class NWAlignment(BaseAlignment):
    """
    Structure for performing and analyzing a Needleman-Wunch global alignment.

    .. note:

        Exposed attributes and methods are the same as ``SSWAlignment``, so
        local and global alignmnts can be handled in the same way. In fact,
        since comparisons are made based on score, local and global alignments
        can be directly compared with constructions like::

            local_aln == global_aln
            local_aln > global_aln
            alignments = sorted([global_aln, local_aln])

    Attributes:

        alignment_type (str): Is 'global' for all ``NWAlignment`` objects.

        aligned_query (str): The aligned query sequence (including gaps).

        aligned_target (str): The aligned target sequence (including gaps).

        alignment_midline (str): Midline for the aligned sequences, with
          ``|`` indicating matches and a gap indicating mismatches::

            print(aln.aligned_query)
            print(aln.alignment_midline)
            print(aln.aligned_target)

            # ATGC
            # || |
            # ATCC

        score (int): Alignment score.

        query_begin (int): Position in the raw query sequence at which
            the optimal alignment begins.

        query_end (int): Position in the raw query sequence at which the
            optimal alignment ends.

        target_begin (int): Position in the raw target sequence at which
            the optimal alignment begins.

        target_end (int): Position in the raw target sequence at which the
            optimal alignment ends.
    """
    def __init__(self, query, target, match=3, mismatch=-2,
        gap_open=-5, gap_extend=-2,
        score_match=None, score_mismatch=None,
        score_gap_open=None, score_gap_extend=None,
        matrix=None, aa=False):
        super(NWAlignment, self).__init__(query, target, matrix,
            match, mismatch, gap_open, gap_extend, aa)
        self.alignment_type = 'global'
        self._score_match = int(score_match) if score_match is not None else None
        self._score_mismatch = int(score_mismatch) if score_mismatch is not None else None
        self._score_gap_open = int(score_gap_open) if score_gap_open is not None else None
        self._score_gap_extend = int(score_gap_extend) if score_gap_extend is not None else None
        self._matrix = matrix
        self._alignment = self._align()
        self.aligned_query = self._alignment[0]
        self.aligned_target = self._alignment[1]
        self.alignment_midline = self._alignment_midline()
        self.score = self._score_alignment()


    def _get_matrix_file(self, match=None, mismatch=None, matrix=None):
        matrix_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'matrices')
        builtins = [os.path.basename(f) for f in list_files(matrix_dir)]
        if matrix is None:
            matrix_name = 'match{}mismatch{}'.format(abs(match), abs(mismatch))
            if matrix_name not in builtins:
                self._build_matrix_from_params(match, mismatch, os.path.join(matrix_dir, matrix_name))
            return os.path.join(matrix_dir, matrix_name)
        elif isinstance(matrix, dict):
            dhash = hashlib.md5()
            encoded = json.dumps(matrix, sort_keys=True).encode()
            dhash.update(encoded)
            matrix_name = dhash.hexdigest()
            if matrix_name not in builtins:
                self._build_matrix_from_dict(matrix, os.path.join(matrix_dir, matrix_name))
            return os.path.join(matrix_dir, matrix_name)
        else:
            if self._matrix.lower() in builtin_names:
                return os.path.join(matrix_dir, self._matrix.lower())
            else:
                err = 'The supplied matrix name ({}) does not exist. '.format(matrix)
                err += 'Built-in matrices are: {}'.format('\n'.join(builtins))
                raise RuntimeError(err)


        # if matrix is not None:
        #     matrix_name = matrix
        # else:
        #     matrix_name = 'match{}mismatch{}'.format(abs(match), abs(mismatch))
        # if matrix_name.lower() in builtins:
        #     return os.path.join(matrix_dir, matrix_name)
        # builtin_names = [os.path.basename(f) for f in list_files(matrix_dir)]
        # if self._matrix is not None:
        #     if self._matrix.lower() in builtin_names:
        #         return os.path.join(matrix_dir, self._matrix.lower())
        #     else:
        #         err = 'The supplied matrix name ({}) does not exist. '.format(matrix)
        #         err += 'Built-in matrices are: {}'.format(', '.join(builtins))
        #         raise RuntimeError(err)
        # else:
        #     self._build_matrix_from_params(match, mismatch, os.path.join(matrix_dir, matrix_name))
        #     return os.path.join(matrix_dir, matrix_name)

    def _align(self):
        matrix = self._get_matrix_file(match=self._match,
                                       mismatch=self._mismatch,
                                       matrix=self._matrix)
        aln = nw.global_align(self.query.sequence,
                              self.target.sequence,
                              gap_open=self._gap_open,
                              gap_extend=self._gap_extend,
                              matrix=matrix)
        return aln

    def _score_alignment(self):
        if all([self._score_match is not None, self._score_mismatch is not None]):
            matrix = self._get_matrix_file(match=self._score_match,
                                           mismatch=self._score_mismatch)
        elif self._matrix is not None:
            matrix = self._get_matrix_file(matrix=self._matrix)
        else:
            matrix = self._get_matrix_file(match=self._match,
                                           mismatch=self._mismatch)
        gap_open = self._score_gap_open if self._score_gap_open is not None else self._gap_open
        gap_extend = self._score_gap_extend if self._score_gap_extend is not None else self._gap_extend
        aln = nw.score_alignment(self.aligned_query,
                                self.aligned_target,
                                gap_open=gap_open,
                                gap_extend=gap_extend,
                                matrix=matrix)
        return aln

    @staticmethod
    def _build_matrix_from_params(match, mismatch, matrix_file):
        mstring = ' {}'.format(match) if len(str(match)) == 1 else str(match)
        mmstring = ' {}'.format(mismatch) if len(str(mismatch)) == 1 else str(mismatch)
        residues = ['A', 'C', 'D', 'E', 'F',
                    'G', 'H', 'I', 'K', 'L',
                    'M', 'N', 'P', 'Q', 'R',
                    'S', 'T', 'V', 'W', 'Y', '*']
        header = '   ' + '  '.join(residues)
        matlist = [header, ]
        for r1 in residues:
            resline = [r1, ]
            for r2 in residues:
                resline.append(mstring if r1 == r2 else mmstring)
            matlist.append(' '.join(resline))
        open(matrix_file, 'w').write('\n'.join(matlist))
        return matrix_file

    @staticmethod
    def _build_matrix_from_dict(matrix, matrix_file):
        residues = sorted(matrix.keys())
        header = '   ' + '  '.join(residues)
        matlist = [header, ]
        for r1 in residues:
            resline = [r1, ]
            for r2 in residues:
                s = matrix[r1][r2]
                score = ' {}'.format(s) if len(str(s)) == 1 else str(s)
                resline.append(score)
            matlist.append(' '.join(resline))
        with open(matrix_file, 'w') as f:
            f.write('\n'.join(matlist))
        return matrix_file

    @staticmethod
    def _get_builtin_matrix(matrix_name):
        matrix_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'matrices')
        matrices = [os.path.basename(f) for f in list_files(matrix_dir)]
        if matrix_name.lower() not in matrices:
            err = 'The maxtrix name you provided ({}) is not built-in.'.format(matrix_name)
            err += 'Built in matrices are: {}'.format(', '.join(matrices))
            raise RuntimeError()
        return os.path.join(matrix_dir, matrix_name.lower())




# -------------------------------------
#
#       VISUAL ALIGNMENT
#
# -------------------------------------



def dot_alignment(sequences, seq_field=None, name_field=None, root=None, root_name=None,
        cluster_threshold=0.75, as_fasta=False, just_alignment=False):
    '''
    Creates a dot alignment (dots indicate identity, mismatches are represented by the mismatched
    residue) for a list of sequences.

    Args:

        sequence (list(Sequence)): A list of Sequence objects to be aligned.

        seq_field (str): Name of the sequence field key. Default is ``vdj_nt``.

        name_field (str): Name of the name field key. Default is ``seq_id``.

        root (str, Sequence): The sequence used to 'root' the alignment. This sequence will be at the
            top of the alignment and is the sequence against which dots (identity) will be evaluated.
            Can be provided either as a string corresponding to the name of one of the sequences in
            ``sequences`` or as a Sequence object. If not provided, ``sequences`` will be clustered
            at ``cluster_threshold`` and the centroid of the largest cluster will be used.

        root_name (str): Name of the root sequence. If not provided, the existing name of the root
            sequence (``name_field``) will be used. If ``root`` is not provided, the default ``root_name``
            is ``'centroid'``.

        cluster_threshold (float): Threshold with which to cluster sequences if ``root`` is not provided.
            Default is ``0.75``.

        as_fasta (bool): If ``True``, returns the dot alignment as a FASTA-formatted string, rather than
            a string formatted for human readability.

        just_alignment (bool): If ``True``, returns just the dot-aligned sequences as a list.

    Returns:

        If ``just_alignment`` is ``True``, a list of dot-aligned sequences (without sequence names) will be returned.
        If ``as_fasta`` is ``True``, a string containing the dot-aligned sequences in FASTA format will be returned.
        Otherwise, a formatted string containing the aligned sequences (with sequence names) will be returned.
    '''
    import abstar

    from .cluster import cluster

    sequences = deepcopy(sequences)
    root = copy(root)

    # if custom seq_field is specified, copy to the .alignment_sequence attribute
    if seq_field is not None:
        if not all([seq_field in list(s.annotations.keys()) for s in sequences]):
            print('\nERROR: {} is not present in all of the supplied sequences.\n'.format(seq_field))
            sys.exit(1)
        for s in sequences:
            s.alignment_sequence = s[seq_field]
    else:
        for s in sequences:
            s.alignment_sequence = s.sequence

    # if custom name_field is specified, copy to the .id attribute
    if name_field is not None:
        if not all([name_field in list(s.annotations.keys()) for s in sequences]):
            print('\nERROR: {} is not present in all of the supplied sequences.\n'.format(name_field))
            sys.exit(1)
        for s in sequences:
            s.alignment_id = s[name_field]
    else:
        for s in sequences:
            s.alignment_id = s.id

    # parse the root sequence
    if all([root is None, root_name is None]):
        clusters = cluster(sequences, threshold=cluster_threshold, quiet=True)
        clusters = sorted(clusters, key=lambda x: x.size, reverse=True)
        centroid = clusters[0].centroid
        root = abstar.run(('centroid', centroid.sequence))
        root.alignment_id = 'centroid'
        root.alignment_sequence = root[seq_field]
    elif type(root) in STR_TYPES:
        root = [s for s in sequences if s.alignment_id == root][0]
        if not root:
            print('\nERROR: The name of the root sequence ({}) was not found in the list of input sequences.'.format(root))
            print('\n')
            sys.exit(1)
        sequences = [s for s in sequences if s.alignment_id != root.alignment_id]
    elif type(root) == Sequence:
        if seq_field is not None:
            if seq_field not in list(root.annotations.keys()):
                print('\nERROR: {} is not present in the supplied root sequence.\n'.format(seq_field))
                sys.exit(1)
            root.alignment_sequence = root[seq_field]
        if name_field is not None:
            if name_field not in list(root.annotations.keys()):
                print('\nERROR: {} is not present in the supplied root sequence.\n'.format(name_field))
                sys.exit(1)
            root.alignment_id = root[name_field]
        sequences = [s for s in sequences if s.alignment_id != root.alignment_id]
    else:
        print('\nERROR: If root is provided, it must be the name of a sequence \
              found in the supplied list of sequences or it must be a Sequence object.')
        print('\n')
        sys.exit(1)

    if root_name is not None:
        root.alignment_id = root_name
    else:
        root_name = root.alignment_id

    # compute and parse the alignment
    seqs = [(root.alignment_id, root.alignment_sequence)]
    seqs += [(s.alignment_id, s.alignment_sequence) for s in sequences]
    aln = muscle(seqs)
    g_aln = [a for a in aln if a.id == root_name][0]
    dots = [(root_name, str(g_aln.seq)), ]
    for seq in [a for a in aln if a.id != root_name]:
        s_aln = ''
        for g, q in zip(str(g_aln.seq), str(seq.seq)):
            if g == q == '-':
                s_aln += '-'
            elif g == q:
                s_aln += '.'
            else:
                s_aln += q
        dots.append((seq.id, s_aln))
    if just_alignment:
            return [d[1] for d in dots]
    name_len = max([len(d[0]) for d in dots]) + 2
    dot_aln = []
    for d in dots:
        if as_fasta:
            dot_aln.append('>{}\n{}'.format(d[0], d[1]))
        else:
            spaces = name_len - len(d[0])
            dot_aln.append(d[0] + ' ' * spaces + d[1])
    return '\n'.join(dot_aln)



import os
import sys

try:
    from setuptools import setup
except ImportError:
    sys.exit('ERROR: setuptools is required.\n')


try: # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
    from pip.req import parse_requirements
# try:
#     from pip.req import parse_requirements
# except ImportError:
#     sys.exit('ERROR: pip is required.\n')


if os.environ.get('READTHEDOCS', None):
    # Set empty install_requires to get install to work on readthedocs
    install_requires = []
else:
    if sys.version_info[0] > 2:
        req_file = 'requirements.txt'
    else:
        req_file = 'requirements2.txt'
    try:
        reqs = parse_requirements(req_file, session=False)
    except TypeError:
        reqs = parse_requirements(req_file)
    try:
        install_requires = [str(r.req) for r in reqs]
    except AttributeError:
        install_requires = [str(r.requirement) for r in reqs]

# read version
exec(open('abutils/version.py').read())

config = {
    'description': 'Utilities for analysis of antibody NGS data',
    'author': 'Bryan Briney',
    'url': 'https://www.github.com/briney/abutils',
    'author_email': 'briney@scripps.edu',
    'version': __version__,
    'install_requires': install_requires,
    'packages': ['abutils'],
    'scripts': ['bin/batch_cellranger'],
    'name': 'abutils',
    'include_package_data': True,
    'classifiers': ['License :: OSI Approved :: MIT License',
                    'Programming Language :: Python :: 3.6',
                    'Programming Language :: Python :: 3.7',
                    'Programming Language :: Python :: 3.8',
                    'Topic :: Scientific/Engineering :: Bio-Informatics']
}

setup(**config)



#!/usr/bin/env python
# filename: io.py


#
# Copyright (c) 2020 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import re

import numpy as np

import scanpy as sc


def read_10x_mtx(mtx_file, gex_only=False, hash_regex='cell ?hash', ignore_hash_regex_case=True,
                 transform_hashes=True, ignore_zero_median_hashes=True,
                 transform_features=True, ignore_zero_median_features=True):
    '''
    Reads a 10x Genomics matrix file and outputs gex, cell hash, and feature barcode data.

    Args:
    -----

        mtx_file (str): path to the 10x Genomics matrix folder (as accepted by ``scanpy.read_10x_mtx()``)

        gex_only (bool): If ``True``, return only gene expression data and ignore features and hashes.
            Default is ``False``.

        hash_regex (str): A regular expression (regex) string used to identify cell hashes. The regex 
            must be found in all hash names. The default is ``'cell ?hash'``, which combined with the
            default setting for ``ignore_hash_regex_case``, will match ``'cellhash'`` or ``'cell hash'``
            in any combination of upper and lower case letters.

        ignore_hash_regex_case (bool): If ``True``, searching for ``hash_regex`` will ignore case.
            Default is ``True``.

        transform_hashes (bool): If ``True``, cell hash counts will be log2 transformed 
            (after adding 1 to the raw count). Default is ``True``.
        
        transform_features (bool): If ``True``, feature counts will be log2 transformed 
            (after adding 1 to the raw count). Default is ``True``.

        ignore_zero_median_hashes (bool): If ``True``, any hashes containing a meadian
            count of ``0`` will be ignored and not returned in the hash dataframe. Default
            is ``True``.
        
        ignore_zero_median_features (bool): If ``True``, any features containing a meadian
            count of ``0`` will be ignored and not returned in the feature dataframe. Default
            is ``True``.
    
    Returns:
    --------

        gex (anndata.AnnData): An ``AnnData`` object containing gene expression data. If ``gex_only``
            is ``True``, only ``gex`` is returned.

        hash_df (pandas.DataFrame): A ``DataFrame`` containing cell hash data. Not returned if 
            ``gex_only`` is ``True``.

        feature_df (pandas.DataFrame): A ``DataFrame`` containing feature data. Not returned if 
            ``gex_only`` is ``True``.
    '''
    # read input file
    adata = sc.read_10x_mtx(mtx_file, gex_only=False)
    # split input files
    gex = adata[:,adata.var.feature_types == 'Gene Expression']
    if gex_only:
        return gex
    non_gex = adata[:,adata.var.feature_types != 'Gene Expression']
    # parse out features and hashes
    if ignore_hash_regex_case:
        pattern = re.compile(hash_regex, flags=re.IGNORECASE)
    else:
        pattern = re.compile(hash_regex)
    hashes = non_gex[:, [re.search(pattern, i) is not None for i in non_gex.var.gene_ids]]
    features = non_gex[:, [re.search(pattern, i) is None for i in non_gex.var.gene_ids]]
    # make hash DataFrame
    hash_df = hashes.to_df()[hashes.var_names]
    if ignore_zero_median_hashes:
        hash_df = hash_df[[h for h in hash_df.columns.values if hash_df[h].median() > 0]]
    if transform_hashes:
        hash_df += 1
        hash_df = hash_df.apply(np.log2)
    # make feature DataFrame
    feature_df = features.to_df()[features.var_names]
    if ignore_zero_median_features:
        feature_df = feature_df[[h for h in feature_df.columns.values if feature_df[h].median() > 0]]
    if transform_features:
        feature_df += 1
        feature_df = feature_df.apply(np.log2)
    return gex, hash_df, feature_df


#!/usr/bin/env python
# filename: log.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os

# from .pipeline import make_dir


def setup_logging(logfile, print_log_location=True, debug=False):
    '''
    Set up logging using the built-in ``logging`` package.

    A stream handler is added to all logs, so that logs at or above
    ``logging.INFO`` level are printed to screen as well as written
    to the log file.

    Arguments:

        logfile (str): Path to the log file. If the parent directory
            does not exist, it will be created. Required.

        print_log_location (bool): If ``True``, the log path will be
            written to the log upon initialization. Default is ``True``.

        debug (bool): If true, the log level will be set to ``logging.DEBUG``.
            If ``False``, the log level will be set to ``logging.INFO``.
            Default is ``False``.
    '''
    log_dir = os.path.dirname(logfile)
    make_dir(log_dir)
    fmt = '[%(levelname)s] %(name)s %(asctime)s %(message)s'
    if debug:
        logging.basicConfig(filename=logfile,
                            filemode='w',
                            format=fmt,
                            level=logging.DEBUG)
    else:
        logging.basicConfig(filename=logfile,
                            filemode='w',
                            format=fmt,
                            level=logging.INFO)
    logger = logging.getLogger('log')
    logger = add_stream_handler(logger)
    if print_log_location:
        logger.info('LOG LOCATION: {}'.format(logfile))


def get_logger(name=None):
    '''
    Get a logging handle.

    As with ``setup_logging``, a stream handler is added to the
    log handle.

    Arguments:

        name (str): Name of the log handle. Default is ``None``.
    '''
    logger = logging.getLogger(name)
    if len(logger.handlers) == 0:
        logger = add_stream_handler(logger)
    return logger


def add_stream_handler(logger):
    formatter = logging.Formatter("%(message)s")
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    ch.setLevel(logging.INFO)
    logger.addHandler(ch)
    return logger


def make_dir(d):
    if not os.path.exists(d):
        os.makedirs(d)
